# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5_agent_auditor.db.connection \
    import Session
from f5_agent_auditor.db import models


class Queries(object):

    def __init__(self):

        self.connection = models.con
        self.lb = models.Loadbalancer
        self.ls = models.Listener
        self.pl = models.Pool
        self.mn = models.Monitor
        self.mb = models.Member
        self.bindings = models.Loadbalanceragentbindings

    def get_loadbalancers_by_agent_id(self, agent_id):
        with Session(self.connection) as se:
            # join is LEFT OUTER JOIN by default
            ret = se.query(self.lb).join(self.bindings).filter(
                  self.bindings.agent_id == agent_id).all()
        return ret

    def get_loadbalancer(self, lb_id):
        with Session(self.connection) as se:
            ret = se.query(self.lb).get(lb_id)
        return ret

    def get_loadbalancers_by_project_id(self, pj_id):
        with Session(self.connection) as se:
            ret = se.query(self.lb).filter(
                self.lb.project_id == pj_id
            ).all()
        return ret

    def get_listener(self, ls_id):
        with Session(self.connection) as se:
            ret = se.query(self.ls).get(ls_id)
        return ret

    def get_listeners_by_lb_id(self, lb_id):
        with Session(self.connection) as se:
            ret = se.query(self.ls).filter(
                self.ls.loadbalancer_id == lb_id
            ).all()
        return ret

    def get_listeners_by_project_id(self, lb_id):
        with Session(self.connection) as se:
            ret = se.query(self.ls).filter(
                self.ls.project_id == lb_id
            ).all()
        return ret

    def get_pool(self, pl_id):
        with Session(self.connection) as se:
            ret = se.query(self.pl).get(pl_id)
        return ret

    def get_pools_by_lb_id(self, lb_id):
        with Session(self.connection) as se:
            ret = se.query(self.pl).filter(
                models.Pool.loadbalancer_id == lb_id
            ).all()
        return ret

    def get_pools_by_project_id(self, pj_id):
        with Session(self.connection) as se:
            ret = se.query(self.pl).filter(
                models.Pool.project_id == pj_id
            ).all()
        return ret

    def get_mn(self, mn_id):
        with Session(self.connection) as se:
            ret = se.query(self.mn).get(mn_id)
        return ret

    def get_member(self, mb_id):
        with Session(self.connection) as se:
            ret = se.query(self.mb).get(mb_id)
        return ret

    def get_members_by_pool_id(self, pl_id):
        with Session(self.connection) as se:
            ret = se.query(self.mb).filter(
                models.Member.pool_id == pl_id
            ).all()
        return ret



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5_agent_auditor.collector import \
    bigip_collector
from f5_agent_auditor.collector import \
    lbaas_collector
from f5_agent_auditor.db import queries
from f5_openstack_agent.lbaasv2.drivers.bigip.service_adapter import \
    ServiceModelAdapter


def get_collectors(collector_type, conf):

    collectors = []
    collector = dict()

    if collector_type == "lbaas":
        db_query = queries.Queries()
        collector['lbaas'] = lbaas_collector.LbassDBCollector(
            db_query, conf.f5_agent)
        collectors.append(collector)

    elif collector_type == "bigip":
        service_adapter = ServiceModelAdapter(conf)
        ips = conf.icontrol_hostname
        hostnames = parse_hostnames(ips)

        for hostname in hostnames:
            source = bigip_collector.BigIPSource(
                hostname, conf.icontrol_username, conf.icontrol_password
            )
            collector[hostname] = bigip_collector.BigIPCollector(
                source.connection,
                service_adapter)
            collectors.append(collector)
            collector = dict()

    else:
        raise Exception("collector type %s is not registered", collector_type)

    return collectors


def parse_hostnames(hostname_str):
    hostnames = hostname_str.split(',')
    hostnames = [item.strip() for item in hostnames]
    return set(hostnames)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5_agent_auditor.db import connection
from f5_agent_auditor import options
from oslo_config import cfg

from sqlalchemy import Column, String, ForeignKey
from sqlalchemy.orm import relationship

options.load_db_options()
options.parse_options()
conf = cfg.CONF

con = connection.Connection(conf)
Base = con.base
metadata = Base.metadata


class Loadbalanceragentbindings(Base):
    __tablename__ = "lbaas_loadbalanceragentbindings"
    __table_args__ = {'autoload': True}

    loadbalancer_id = Column(
        String, ForeignKey('lbaas_loadbalancers.id'),
        primary_key=True)
    agent_id = Column(String)


class Loadbalancer(Base):
    __tablename__ = 'lbaas_loadbalancers'
    __table_args__ = {'autoload': True, 'extend_existing': True}

    project_id = Column("project_id")
    provisioning_status = Column("provisioning_status")


class Listener(Base):
    __tablename__ = 'lbaas_listeners'
    __table_args__ = {'autoload': True}

    loadbalancer_id = Column("loadbalancer_id")
    provisioning_status = Column("provisioning_status")


class Pool(Base):
    __tablename__ = 'lbaas_pools'
    __table_args__ = {'autoload': True, 'extend_existing': True}

    loadbalancer_id = Column("loadbalancer_id")
    healthmonitor_id = Column(String, ForeignKey('lbaas_healthmonitors.id'))

    # use eagerly load
    members = relationship("Member", lazy='subquery')
    healthmonitor = relationship("Monitor", uselist=False, lazy='subquery')

    provisioning_status = Column("provisioning_status")


class Monitor(Base):
    __tablename__ = 'lbaas_healthmonitors'
    __table_args__ = {'autoload': True}


class Member(Base):
    __tablename__ = 'lbaas_members'
    __table_args__ = {'autoload': True}

    pool_id = Column(String, ForeignKey('lbaas_pools.id'))

    provisioning_status = Column("provisioning_status")



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class LbaasToBigIP(object):

    def __init__(self, benchmark, benchmark_filter):

        self.benchmark_name = None
        self.benchmark = None
        self.benchmark_filter = None
        self.benchmark_projects = None

        self.subject_name = None
        self.subject = None
        self.subject_filter = None
        self.subject_projects = None

        self.validate_subject(benchmark)
        self.init_benchmark(benchmark, benchmark_filter)

    def compare_to(self, subject, subject_filter):
        self.validate_subject(subject)
        self.init_subject(subject, subject_filter)

    def validate_subject(self, subject):
        if not isinstance(subject, dict):
            raise Exception("Comparator must be a dcit type")
        if len(subject) != 1:
            raise Exception("Only one Comparator should be "
                            "provided at a time")

    def init_subject(self, subject, subject_filter):
        self.subject_name = subject.keys()[0]
        self.subject = subject.values()[0]

        self.subject_filter = subject_filter
        projects = self.subject.get_projects_on_device()
        self.subject_projects = self.subject_filter.get_ids(
            projects
        )

    def init_benchmark(self, benchmark, benchmark_filter):
        self.benchmark_name = benchmark.keys()[0]
        self.benchmark = benchmark.values()[0]
        self.benchmark_filter = benchmark_filter

        projects = \
            self.benchmark.get_projects_on_device()
        self.benchmark_projects = set(projects)

    def get_common_resources_diff(self, bm_method,
                                  sub_method,
                                  resource_type=None):
        bm_resources = []
        sub_resources = []

        for project in self.benchmark_projects:
            bm_resources += bm_method(
                project
            )

        bm_res = self.benchmark_filter.get_resources(
            bm_resources)
        bm_ids = set(bm_res.keys())

        for project in self.subject_projects:
            sub_resources += sub_method(
                project
            )

        sub_ids = self.subject_filter.get_ids(
            sub_resources)

        diff = bm_ids - sub_ids

        result = self.benchmark_filter.convert_common_resources(
            diff, bm_res, resource_type=resource_type
        )

        return result

    def get_missing_projects(self):
        res = self.benchmark_projects - self.subject_projects
        diff = self.benchmark_filter.convert_projects(
            res
        )
        return diff

    def get_missing_loadbalancers(self):
        bm_method = self.benchmark.get_project_loadbalancers
        sub_method = self.subject.get_project_loadbalancers
        diff = self.get_common_resources_diff(
            bm_method, sub_method, "loadbalancer"
        )
        return diff

    def get_missing_listeners(self):
        bm_method = self.benchmark.get_project_listeners
        sub_method = self.subject.get_project_listeners
        diff = self.get_common_resources_diff(
            bm_method, sub_method, "listener"
        )
        return diff

    def get_missing_pools(self):
        bm_method = self.benchmark.get_project_pools
        sub_method = self.subject.get_project_pools
        diff = self.get_common_resources_diff(
            bm_method, sub_method, "pool"
        )
        return diff

    def get_missing_members(self):
        bm_pools = []
        sub_pools = []
        missing_mb = []

        for project in self.benchmark_projects:
            bm_pools += self.benchmark.get_project_pools(
                project
            )

        bm_mbs = self.benchmark_filter.filter_pool_members(bm_pools)

        for project in self.subject_projects:
            sub_pools += self.subject.get_project_pools(
                project
            )

        sub_mbs = self.subject_filter.filter_pool_members(sub_pools)

        for pool_id, members in bm_mbs.items():
            if pool_id not in sub_mbs:
                if members:
                    missing_mb += self.benchmark_filter.convert_members(
                        pool_id, members)
                continue

            for mb in members:
                if not mb["address_port"] in sub_mbs[pool_id]:
                    missing_mb += self.benchmark_filter.convert_members(
                        pool_id, [mb])

        return missing_mb



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from enum import Enum
from f5_agent_auditor.utils import \
    get_filter

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class ResourceType(Enum):
    u"""Defines supported BIG-IP resource types."""

    nat = 1
    pool = 2
    sys = 3
    virtual = 4
    member = 5
    folder = 6
    http_monitor = 7
    https_monitor = 8
    tcp_monitor = 9
    ping_monitor = 10
    node = 11
    snat = 12
    snatpool = 13
    snat_translation = 14
    selfip = 15
    rule = 16
    vlan = 17
    arp = 18
    route_domain = 19
    tunnel = 20
    virtual_address = 21
    l7policy = 22
    client_ssl_profile = 23
    server_ssl_profile = 24
    tcp_profile = 25
    persistence = 26
    cookie_persistence = 27
    dest_addr_persistence = 28
    hash_persistence = 29
    msrdp_persistence = 30
    sip_persistence = 31
    source_addr_persistence = 32
    ssl_persistence = 33
    universal_persistence = 34
    ssl_cert_file = 35
    http_profile = 36
    oneconnect = 37
    udp_monitor = 38
    sip_monitor = 39
    diameter_monitor = 40
    ftp_profile = 41
    partition = 42


class BigIPResourceHelper(object):
    u"""Helper class for creating, updating and deleting BIG-IP resources.

    Reduces some of the boilerplate that surrounds using the F5 SDK.
    Example usage:
        bigip = BigIP("10.1.1.1", "admin", "admin")
        pool = {"name": "pool1",
                "partition": "Common",
                "description": "Default pool",
                "loadBalancingMode": "round-robin"}
        pool_helper = BigIPResourceHelper(ResourceType.pool)
        p = pool_helper.create(bigip, pool)
    """

    def __init__(self, resource_type):
        """Initialize a resource helper."""
        self.resource_type = resource_type

    def create(self, bigip, model):
        u"""Create/update resource (e.g., pool) on a BIG-IP system.

        First checks to see if resource has been created and creates
        it if not.

        :param bigip: BigIP instance to use for creating resource.
        :param model: Dictionary of BIG-IP attributes to add resource. Must
        include name and partition.
        :returns: created or updated resource object.
        """
        resource = self._resource(bigip)
        obj = resource.create(**model)

        return obj

    def exists(self, bigip, name=None, partition=None):
        """Test for the existence of a resource."""
        resource = self._resource(bigip)
        return resource.exists(name=name, partition=partition)

    def delete(self, bigip, name=None, partition=None):
        u"""Delete a resource on a BIG-IP system.

        Checks if resource exists and deletes it. Returns without error
        if resource does not exist.

        :param bigip: BigIP instance to use for creating resource.
        :param name: Name of resource to delete.
        :param partition: Partition name for resou
        """
        resource = self._resource(bigip)
        if resource.exists(name=name, partition=partition):
            obj = resource.load(name=name, partition=partition)
            obj.delete()

    def load(self, bigip, name=None, partition=None):
        u"""Retrieve a BIG-IP resource from a BIG-IP.

        Populates a resource object with attributes for instance on a
        BIG-IP system.

        :param bigip: BigIP instance to use for creating resource.
        :param name: Name of resource to load.
        :param partition: Partition name for resource.
        :returns: created or updated resource object.
        """
        resource = self._resource(bigip)
        return resource.load(name=name, partition=partition)

    def update(self, bigip, model):
        u"""Update a resource (e.g., pool) on a BIG-IP system.

        Modifies a resource on a BIG-IP system using attributes
        defined in the model object.

        :param bigip: BigIP instance to use for creating resource.
        :param model: Dictionary of BIG-IP attributes to update resource.
        Must include name and partition in order to identify resource.
        """
        partition = None
        if "partition" in model:
            partition = model["partition"]
        resource = self.load(bigip, name=model["name"], partition=partition)
        resource.modify(**model)

        return resource

    def get_resources(self, bigip, partition=None,
                      expand_subcollections=False):
        u"""Retrieve a collection BIG-IP of resources from a BIG-IP.

        Generates a list of resources objects on a BIG-IP system.

        :param bigip: BigIP instance to use for creating resource.
        :param name: Name of resource to load.
        :param partition: Partition name for resource.
        :returns: list of created or updated resource objects.
        """
        resources = []
        try:
            collection = self._collection(bigip)
        except KeyError as err:
            LOG.exception(err.message)
            raise err

        if collection:
            params = {'params': ''}
            if partition:
                params['params'] = get_filter(
                    bigip, 'partition', 'eq', partition)
                if expand_subcollections and \
                        isinstance(params['params'], dict):
                    params['params']['expandSubcollections'] = 'true'
                elif expand_subcollections:
                    params['params'] += '&expandSubCollections=true'
                resources = collection.get_collection(requests_params=params)
            else:
                resources = collection.get_collection()

        return resources

    def exists_in_collection(self, bigip, name, partition='Common'):
        collection = self.get_resources(bigip, partition='Common')
        for item in collection:
            if item.name == name:
                return True

        return False

    def _resource(self, bigip):
        return {
            ResourceType.nat: lambda bigip: bigip.tm.ltm.nats.nat,
            ResourceType.pool: lambda bigip: bigip.tm.ltm.pools.pool,
            ResourceType.sys: lambda bigip: bigip.tm.sys,
            ResourceType.virtual: lambda bigip: bigip.tm.ltm.virtuals.virtual,
            ResourceType.member: lambda bigip: bigip.tm.ltm.pools.pool.member,
            ResourceType.folder: lambda bigip: bigip.tm.sys.folders.folder,
            ResourceType.http_monitor:
                lambda bigip: bigip.tm.ltm.monitor.https.http,
            ResourceType.https_monitor:
                lambda bigip: bigip.tm.ltm.monitor.https_s.https,
            ResourceType.tcp_monitor:
                lambda bigip: bigip.tm.ltm.monitor.tcps.tcp,
            ResourceType.udp_monitor:
                lambda bigip: bigip.tm.ltm.monitor.udps.udp,
            ResourceType.sip_monitor:
                lambda bigip: bigip.tm.ltm.monitor.sips.sip,
            ResourceType.diameter_monitor:
                lambda bigip: bigip.tm.ltm.monitor.diameters.diameter,
            ResourceType.ping_monitor:
                lambda bigip: bigip.tm.ltm.monitor.gateway_icmps.gateway_icmp,
            ResourceType.node: lambda bigip: bigip.tm.ltm.nodes.node,
            ResourceType.snat: lambda bigip: bigip.tm.ltm.snats.snat,
            ResourceType.snatpool:
                lambda bigip: bigip.tm.ltm.snatpools.snatpool,
            ResourceType.snat_translation:
                lambda bigip: bigip.tm.ltm.snat_translations.snat_translation,
            ResourceType.selfip:
                lambda bigip: bigip.tm.net.selfips.selfip,
            ResourceType.rule:
                lambda bigip: bigip.tm.ltm.rules.rule,
            ResourceType.vlan:
                lambda bigip: bigip.tm.net.vlans.vlan,
            ResourceType.arp:
                lambda bigip: bigip.tm.net.arps.arp,
            ResourceType.route_domain:
                lambda bigip: bigip.tm.net.route_domains.route_domain,
            ResourceType.tunnel:
                lambda bigip: bigip.tm.net.tunnels.tunnels.tunnel,
            ResourceType.virtual_address:
                lambda bigip: bigip.tm.ltm.virtual_address_s.virtual_address,
            ResourceType.l7policy:
                lambda bigip: bigip.tm.ltm.policys.policy,
            ResourceType.client_ssl_profile:
                lambda bigip: bigip.tm.ltm.profile.client_ssls.client_ssl,
            ResourceType.server_ssl_profile:
                lambda bigip: bigip.tm.ltm.profile.server_ssls.server_ssl,
            ResourceType.tcp_profile:
                lambda bigip: bigip.tm.ltm.profile.tcps.tcp,
            ResourceType.persistence:
                lambda bigip: bigip.tm.ltm.persistence,
            ResourceType.cookie_persistence:
                lambda bigip: bigip.tm.ltm.persistence.cookies.cookie,
            ResourceType.dest_addr_persistence:
                lambda bigip: bigip.tm.ltm.persistence.dest_addrs.dest_addr,
            ResourceType.hash_persistence:
                lambda bigip: bigip.tm.ltm.persistence.hashs.hash,
            ResourceType.msrdp_persistence:
                lambda bigip: bigip.tm.ltm.persistence.msrdp,
            ResourceType.sip_persistence:
                lambda bigip: bigip.tm.ltm.persistence.sips,
            ResourceType.source_addr_persistence:
                lambda bigip:
                    bigip.tm.ltm.persistence.source_addrs.source_addr,
            ResourceType.ssl_persistence:
                lambda bigip: bigip.tm.ltm.persistence.ssl,
            ResourceType.universal_persistence:
                lambda bigip: bigip.tm.ltm.persistence.universal,
            ResourceType.http_profile:
                lambda bigip: bigip.tm.ltm.profile.https.http,
            ResourceType.ssl_cert_file:
                lambda bigip: bigip.tm.sys.file.ssl_certs.ssl_cert,
            ResourceType.ftp_profile:
                lambda bigip: bigip.tm.ltm.profile.ftps.ftp
        }[self.resource_type](bigip)

    def _collection(self, bigip):
        collection_map = {
            ResourceType.nat: lambda bigip: bigip.tm.ltm.nats,
            ResourceType.pool: lambda bigip: bigip.tm.ltm.pools,
            ResourceType.sys: lambda bigip: bigip.tm.sys,
            ResourceType.virtual: lambda bigip: bigip.tm.ltm.virtuals,
            ResourceType.member: lambda bigip: bigip.tm.ltm.pools.pool.member,
            ResourceType.folder: lambda bigip: bigip.tm.sys.folders,
            ResourceType.http_monitor:
                lambda bigip: bigip.tm.ltm.monitor.https,
            ResourceType.https_monitor:
                lambda bigip: bigip.tm.ltm.monitor.https_s,
            ResourceType.tcp_monitor:
                lambda bigip: bigip.tm.ltm.monitor.tcps,
            ResourceType.ping_monitor:
                lambda bigip: bigip.tm.ltm.monitor.gateway_icmps,
            ResourceType.node: lambda bigip: bigip.tm.ltm.nodes,
            ResourceType.snat: lambda bigip: bigip.tm.ltm.snats,
            ResourceType.snatpool:
                lambda bigip: bigip.tm.ltm.snatpools,
            ResourceType.snat_translation:
                lambda bigip: bigip.tm.ltm.snat_translations,
            ResourceType.selfip:
                lambda bigip: bigip.tm.net.selfips,
            ResourceType.rule:
                lambda bigip: bigip.tm.ltm.rules,
            ResourceType.route_domain:
                lambda bigip: bigip.tm.net.route_domains,
            ResourceType.vlan:
                lambda bigip: bigip.tm.net.vlans,
            ResourceType.arp:
                lambda bigip: bigip.tm.net.arps,
            ResourceType.tunnel:
                lambda bigip: bigip.tm.net.tunnels.tunnels,
            ResourceType.virtual_address:
                lambda bigip: bigip.tm.ltm.virtual_address_s,
            ResourceType.l7policy:
                lambda bigip: bigip.tm.ltm.policys,
            ResourceType.client_ssl_profile:
                lambda bigip: bigip.tm.ltm.profile.client_ssls,
            ResourceType.server_ssl_profile:
                lambda bigip: bigip.tm.ltm.profile.server_ssls,
            ResourceType.tcp_profile:
                lambda bigip: bigip.tm.ltm.profile.tcps,
            ResourceType.persistence:
                lambda bigip: bigip.tm.ltm.persistence,
            ResourceType.cookie_persistence:
                lambda bigip: bigip.tm.ltm.persistence.cookies,
            ResourceType.dest_addr_persistence:
                lambda bigip: bigip.tm.ltm.persistence.dest_addrs,
            ResourceType.hash_persistence:
                lambda bigip: bigip.tm.ltm.persistence.hashs,
            ResourceType.msrdp_persistence:
                lambda bigip: bigip.tm.ltm.persistence.msrdps,
            ResourceType.sip_persistence:
                lambda bigip: bigip.tm.ltm.persistence.sips,
            ResourceType.source_addr_persistence:
                lambda bigip: bigip.tm.ltm.persistence.source_addrs,
            ResourceType.ssl_persistence:
                lambda bigip: bigip.tm.ltm.persistence.ssls,
            ResourceType.universal_persistence:
                lambda bigip: bigip.tm.ltm.persistence.universals,
            ResourceType.ssl_cert_file:
                lambda bigip: bigip.tm.sys.file.ssl_certs,
            ResourceType.http_profile:
                lambda bigip: bigip.tm.ltm.profile.https,
            ResourceType.oneconnect:
                lambda bigip: bigip.tm.ltm.profile.one_connects,
            ResourceType.partition:
                lambda bigip: bigip.tm.auth.partitions
        }

        if self.resource_type in collection_map:
            return collection_map[self.resource_type](bigip)
        else:
            LOG.error("Error attempting to get collection for "
                      "resource %s", self.resource_type)
            raise KeyError("No collection available for %s" %
                           (self.resource_type))

    def get_stats(self, bigip, name=None, partition=None, stat_keys=[]):
        """Returns dictionary of stats.

        Use by calling with an array of stats to get from resource. Return
        value will be a dict with key/value pairs. The stat key will only
        be included in the return dict if the resource includes that stat.

        :param bigip: BIG-IP to get stats from.
        :param name: name of resource object.
        :param partition: partition where to get resource.
        :param stat_keys: Array of strings that define stats to collect.
        :return: dictionary with key/value pairs where key is string
        defined in input array, if present in resource stats, and value
        as the value of resource stats 'value' key.
        """
        collected_stats = {}

        # get resource, then its stats
        if self.exists(bigip, name=name, partition=partition):
            resource = self.load(bigip, name=name, partition=partition)
            collected_stats = self.collect_stats(resource, stat_keys)

        return collected_stats

    def collect_stats(self, resource, stat_keys=[]):
        collected_stats = {}
        resource_stats = resource.stats.load()
        stat_entries = resource_stats.entries

        # Difference between 11.6 and 12.1. Stats in 12.1 are embedded
        # in nestedStats. In 11.6, they are directly accessible in entries.
        if stat_keys[0] not in stat_entries:
            # find nestedStats
            for key in stat_entries.keys():
                value = stat_entries.get(key, None)
                if 'nestedStats' in value:
                    stat_entries = value['nestedStats']['entries']

        # add stats defined in input stats array
        for stat_key in stat_keys:
            if stat_key in stat_entries:
                if 'value' in stat_entries[stat_key]:
                    collected_stats[stat_key] = stat_entries[stat_key][
                        'value']
                elif 'description' in stat_entries[stat_key]:
                    collected_stats[stat_key] = \
                        stat_entries[stat_key]['description']

        return collected_stats



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from distutils.version import LooseVersion
import time


def time_logger(logger):
    def timer(func):
        def wrapper(*args, **kwargs):
            start = time.time()
            result = func(*args, **kwargs)
            end = time.time()
            elapse = end - start
            logger.info(
                "%s takes %s seconds",
                func.__name__, elapse
            )
            return result
        return wrapper
    return timer


def get_filter(bigip, key, op, value):
    if LooseVersion(bigip.tmos_version) < LooseVersion('11.6.0'):
        return '$filter=%s+%s+%s' % (key, op, value)
    else:
        return {'$filter': '%s %s %s' % (key, op, value)}



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5_agent_auditor.collector \
    import base
from f5_agent_auditor.utils \
    import time_logger
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class LbassDBCollector(base.Collector):

    def __init__(self, source, agent_id):
        self.source = source
        self.agent_id = agent_id

        # NOTE(pzhang): should not construct cache here
        # cache content {'project_uuid':[resource_a, ...]}
        # eager load
        self.lb_cache = {}
        self.project_cache = []

        # these cache are lazy load
        self.ls_cache = {}
        self.pl_cache = {}

        self.init_lb_cache(self.agent_id)
        self.init_agent_projects()

    def _cache_agent_lb_by_project_id(self,
                                      lbs,
                                      lb_cache):
        for lb in lbs:
            if lb.project_id not in lb_cache:
                self.lb_cache[lb.project_id] = [lb]
            else:
                self.lb_cache[lb.project_id].append(lb)

    def _cache_resource_by_project_id(self,
                                      project_id,
                                      resources,
                                      resource_cache):
        # new search overwrite all old data
        resource_cache[project_id] = resources

    def init_lb_cache(self, agent_id):
        """Initiate root loadbalancer cache for agent

        this is should not be lazy load
        agent_id:str is passed by users
        """
        lbs = self.source.get_loadbalancers_by_agent_id(
            agent_id)
        self._cache_agent_lb_by_project_id(lbs,
                                           self.lb_cache)

    def init_agent_projects(self):
        self.project_cache = self.lb_cache.keys()

    @time_logger(LOG)
    def get_projects_on_device(self):
        LOG.info("Get projects of agent : %s in Neutron DB",
                 self.agent_id)
        return self.project_cache

    @time_logger(LOG)
    def get_project_loadbalancers(self, project_id):
        LOG.info("Get loadbalancers of project: %s", project_id)
        if project_id not in self.project_cache:
            # project is not exist in cache
            return []

        # assert or throw error
        loadbalancers = self.lb_cache[project_id]
        return loadbalancers

    @time_logger(LOG)
    def get_project_listeners(self, project_id):
        LOG.info("Get listeners of project: %s", project_id)
        total_listeners = []

        if project_id not in self.project_cache:
            # project is not exist in cache
            return []

        if project_id in self.ls_cache:
            # find ls of project in cache
            return self.ls_cache[project_id]

        total_listeners = self.source.get_listeners_by_project_id(
           project_id
        )

        self._cache_resource_by_project_id(project_id,
                                           total_listeners,
                                           self.ls_cache)

        return total_listeners

    @time_logger(LOG)
    def get_project_pools(self, project_id):
        LOG.info("Get pools of project: %s", project_id)
        total_pools = []

        if project_id not in self.project_cache:
            # project is not exist in cache
            return []

        if project_id in self.pl_cache:
            pools = self.pl_cache[project_id]
            return pools

        total_pools = self.source.get_pools_by_project_id(
            project_id
        )

        self._cache_resource_by_project_id(project_id,
                                           total_pools,
                                           self.pl_cache)

        # set pl_cache first, then set member
        # or it will be in recursive loop
        # if cache do not have any pools belongs to project_id
        # , it will cause sink into infinite recursive.
        self.set_project_pool_members(project_id)

        return total_pools

    @time_logger(LOG)
    def set_project_pool_members(self, project_id):
        """Set members in a pool of a project

           consider members are as configuration of pools
        """
        LOG.info("Set pool members of project: %s", project_id)
        if project_id not in self.project_cache:
            # project is not exist in cache
            return

        pools = self.get_project_pools(project_id)
        for pl in pools:
            pool_configured_members = []
            pool_configured_members = self.source.get_members_by_pool_id(pl.id)
            # maybe use it this way: pl["members"]
            pl.members = pool_configured_members



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5_agent_auditor.collector import factory
from f5_agent_auditor import comparator
from f5_agent_auditor.filters import BigIPFilter
from f5_agent_auditor.filters import LbaasFilter
from f5_agent_auditor import options
from f5_agent_auditor.publishers.csv_publisher import \
    CSVPublisher
from f5_agent_auditor.utils \
    import time_logger

from oslo_config import cfg
from oslo_log import log as logging


options.load_options()
options.parse_options()
conf = cfg.CONF

logging.setup(conf, __name__)
LOG = logging.getLogger(__name__)


@time_logger(LOG)
def main():

    if conf.f5_agent:
        bigip_filter = BigIPFilter(conf.environment_prefix)
        lbaas_filter = LbaasFilter()

        lbaas_collector = factory.get_collectors("lbaas", conf)[0]
        bigip_collectors = factory.get_collectors("bigip", conf)

        comp = comparator.LbaasToBigIP(lbaas_collector, lbaas_filter)

        for collector in bigip_collectors:
            comp.compare_to(collector, bigip_filter)

            hostname = collector.keys()[0]
            csv_publisher = CSVPublisher()
            csv_publisher.set_filepath(hostname)
            csv_publisher.set_csv_fields(
                "resource type", "uuid",
                "provisioning status", "project id",
                "pool id", "detail"
            )

            missing = []
            missing += comp.get_missing_projects()
            missing += comp.get_missing_loadbalancers()
            missing += comp.get_missing_listeners()
            missing += comp.get_missing_pools()
            missing += comp.get_missing_members()

            if missing:
                csv_publisher.publish(*missing)

    else:
        raise Exception("Provide an corresponding agent ID "
                        "--f5-agent")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class BigIPFilter(object):

    def __init__(self, prefix):
        self.prefix = prefix + "_"

    def get_id(self, resource):
        uuid = resource.name.split(self.prefix)[-1]
        return uuid

    def get_ids(self, resources):
        ids = []
        for res in resources:
            if self.prefix in res.name:
                uuid = self.get_id(res)
                ids.append(uuid)
        return set(ids)

    @staticmethod
    def format_member(member):
        address = None
        port = None
        mb = member.get('name')

        if "%" in mb:
            mb = mb.split("%")
            address = mb[0]
            if "." in mb[1]:
                port = mb[1].split('.')[1]
            else:
                port = mb[1].split(":")[1]
        else:
            if mb.count(":") > mb.count("."):
                mb = mb.split('.')
            else:
                mb = mb.split(':')
            address = mb[0]
            port = mb[1]
        address_port = address + "_" + str(port)
        return address_port

    def filter_pool_members(self, partition_pools):
        pools = {}
        for pl in partition_pools:
            members = []
            pl_id = None
            member_items = pl.membersReference.get('items')
            if member_items:
                for mb in member_items:
                    member = self.format_member(mb)
                    members.append(member)
            pl_id = self.get_id(pl)
            pools[pl_id] = members
        return pools


class LbaasFilter(object):

    def get_resources(self, resources):
        result = {}
        for res in resources:
            result[res.id] = res
        return result

    @staticmethod
    def format_member(member):
        mb = {}
        mb['address_port'] = member.address + "_" + str(member.protocol_port)
        mb['id'] = member.id
        mb['provisioning_status'] = member.provisioning_status
        mb['project_id'] = member.project_id
        return mb

    def filter_pool_members(self, project_pools):
        pools = {}
        for pl in project_pools:
            members = []
            for mb in pl.members:
                member = self.format_member(mb)
                members.append(member)
            pools[pl.id] = members
        return pools

    def convert_projects(self, diff_ids):
        res = [
            {
                "resource type": "project",
                "uuid": uuid,
            }
            for uuid in diff_ids
        ]
        return res

    def convert_common_resources(self,
                                 diff_ids,
                                 resources,
                                 resource_type=None):
        res = [
            {
                "resource type": resource_type,
                "uuid": uuid,
                "provisioning status": resources[
                    uuid].provisioning_status,
                "project id": resources[
                    uuid].project_id
            }
            for uuid in diff_ids
        ]
        return res

    def convert_members(self, pool_id, members):
        mbs = [{
            "resource type": "member",
            "uuid": m['id'],
            "provisioning status": m['provisioning_status'],
            "project id": m["project_id"],
            "pool id": pool_id,
            "detail": {
                "address": m['address_port'].split("_")[0],
                "port": m['address_port'].split("_")[1]
            }
        } for m in members]

        return mbs



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_config import cfg
import sys

# require f5 agent is installed
import f5_openstack_agent.lbaasv2.drivers.bigip.agent_manager as manager
from f5_openstack_agent.lbaasv2.drivers.bigip import icontrol_driver
from oslo_db import options

tool_opts = [
    cfg.StrOpt("f5-agent",
               short="ag",
               default=None,
               help=("Provide an ID of an agent"))
]

cfg.CONF.register_cli_opts(tool_opts)


def load_options(conf=cfg.CONF):
    conf.register_opts(manager.OPTS)
    conf.register_opts(icontrol_driver.OPTS)


def load_db_options(conf=cfg.CONF):
    options.set_defaults(conf)


def parse_options(args=sys.argv[1:],
                  conf=cfg.CONF,
                  project="f5-agent-auditor"):
    conf(args, project)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session

import eventlet
eventlet.monkey_patch(thread=True)


def get_engine(conf):

    engine_args = {
        "pool_recycle": conf.database.idle_timeout,
        "echo": False,
        "pool_size": conf.database.max_pool_size,
        "pool_timeout": conf.database.pool_timeout
    }

    return create_engine(conf.database.connection, **engine_args)


class Connection(object):

    __instance = None

    def __new__(cls, conf):
        if not isinstance(cls.__instance, cls):
            cls.__instance = super(Connection, cls).__new__(cls)
            cls.__instance.engine = get_engine(conf)
            cls.__instance.base = declarative_base(cls.__instance.engine)
        return cls.__instance


class Session(object):

    def __init__(self, conn):
        self.connection = conn
        self.engine = self.connection.engine
        self.scope_session = scoped_session(sessionmaker(
            bind=self.connection.engine))

    def __enter__(self):
        # print(self.engine.pool.status())
        session = self.scope_session()
        return session

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.scope_session.remove()



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from f5.bigip import ManagementRoot
from f5_agent_auditor.collector import base
from f5_agent_auditor import resource_helper
from f5_agent_auditor.utils import time_logger
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class BigIPSource(object):
    def __init__(self, hostname,
                 icontrol_username,
                 icontrol_password):
        self.__hostname = hostname
        self.__icontrol_username = icontrol_username
        self.__icontrol_password = icontrol_password

        self.connection = self.init_bigip()

    @property
    def hostname(self):
        return self.__hostname

    @property
    def username(self):
        return self.__icontrol_username

    @property
    def password(self):
        return None

    def init_bigip(self):
        bigip = ManagementRoot(self.__hostname,
                               self.__icontrol_username,
                               self.__icontrol_password)
        return bigip


class BigIPCollector(base.Collector):
    def __init__(self, source, service_adapter):
        self.bigip = source
        self.service_adapter = service_adapter

        self.partition_helper = resource_helper.BigIPResourceHelper(
            resource_helper.ResourceType.partition)
        self.vip_helper = resource_helper.BigIPResourceHelper(
            resource_helper.ResourceType.virtual_address)
        self.vs_helper = resource_helper.BigIPResourceHelper(
            resource_helper.ResourceType.virtual)
        self.pool_helper = resource_helper.BigIPResourceHelper(
            resource_helper.ResourceType.pool)

    @staticmethod
    def convert_member_name(bigip_member_name):
        # remove the routedomain id
        if "%" in bigip_member_name:
            port = bigip_member_name.split(":")[1]
            address = bigip_member_name.split("%")[0]
            bigip_member_name = address + ":" + port

        return bigip_member_name

    @time_logger(LOG)
    def get_projects_on_device(self):
        LOG.info("Get projects on device %s", self.bigip.hostname)
        partitions = self.partition_helper.get_resources(self.bigip)
        return partitions

    @time_logger(LOG)
    def get_project_loadbalancers(self, project_id):
        LOG.info("Get loadbalancers of project: %s", project_id)
        folder_name = self.service_adapter.get_folder_name(project_id)
        loadbalancers = self.vip_helper.get_resources(self.bigip, folder_name)
        return loadbalancers

    @time_logger(LOG)
    def get_project_listeners(self, project_id):
        LOG.info("Get listeners of project: %s", project_id)
        folder_name = self.service_adapter.get_folder_name(project_id)
        listeners = self.vs_helper.get_resources(self.bigip, folder_name)
        return listeners

    @time_logger(LOG)
    def get_project_pools(self, project_id):
        LOG.info("Get pools of project: %s", project_id)
        folder_name = self.service_adapter.get_folder_name(project_id)
        pools = self.pool_helper.get_resources(self.bigip, folder_name, True)
        return pools



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import csv
import os


class CSVPublisher(object):

    def __init__(self):
        self.filepath = None

    def set_filepath(self, filename):
        self.filepath = "/tmp/check_" + str(
           filename) + ".csv"

    def set_csv_fields(self, *args):
        self.fields = args

    def publish(self, *args):
        if os.path.isfile(self.filepath):
            os.remove(self.filepath)
        with open(self.filepath, mode='a') as csv_file:
            if self.fields:
                publisher = csv.DictWriter(
                    csv_file, fieldnames=self.fields)
                publisher.writeheader()
            else:
                publisher = csv.writer(
                    csv_file,
                    delimiter=',',
                    quotechar='"',
                    quoting=csv.QUOTE_MINIMAL)
            for res in args:
                publisher.writerow(res)



# -*- coding: utf-8 -*-

import setuptools

setuptools.setup(
    install_requires=['f5-openstack-agent>=9.6.0'],
    # this works, but packages in setup.cfg not work
    packages=setuptools.find_packages()
)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import os
import shutil


class FilePublisher(object):

    def __init__(self, dirctory):
        self.path = "/tmp/check_" + str(dirctory)
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)
        os.mkdir(self.path)

    def publish(self, filename, context):
        file_path = self.path + "/" + filename
        with open(file_path, 'a') as f:
            json.dump(context, f, indent=2, sort_keys=True)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class Collector(object):

    def __init__(self, source):
        self.source = source

    def get_project_loadbalancers(project_id, *args, **kwargs):
        """Get loadbalancers by a project_id.

        """
        pass

    def get_project_listeners(project_id, *args, **kwargs):
        """Get listeners by a project_id.

        """
        pass

    def get_project_pools(project_id, *args, **kwargs):
        """Get pools by a project_id.

        """
        pass

    def get_project_members(project_id, *args, **kwargs):
        """Get members by a project_id.

        """
        pass

    def get_project_healthmonitor(project_id, *args, **kwargs):
        """Get healthmonitor by a project_id.

        """
        pass
